"插件市场客户端"
import os
import platform
import shutil
import tempfile
import traceback
import time
import shlex
import requests
import ujson as json
from tooldelta import urlmethod
from .utils import Utils
from .color_print import Print
from .plugin_load import PluginRegData
from .cfg import Cfg
from .constants import (
    PLUGIN_MARKET_SOURCE_OFFICIAL,
    TOOLDELTA_CLASSIC_PLUGIN,
    TOOLDELTA_INJECTED_PLUGIN
)
from .plugin_load.PluginGroup import plugin_group

if platform.system().lower() == "windows":
    CLS_CMD = "cls"
else:
    CLS_CMD = "clear"


def clear_screen() -> None:
    "清屏"
    os.system(shlex.quote(CLS_CMD))


def path_dir(path: str) -> str | None:
    """获取路径的上一级目录

    Args:
        path (str): 路径

    Returns:
        str|None: 上一级目录
    """
    if "/" not in path:
        return None
    return "/".join(path.split("/")[:-1])


def url_join(*urls) -> str:
    """连接URL

    Returns:
        str: 连接后的URL
    """
    return "/".join(urls)


def get_json_from_url(url: str) -> dict:
    """从URL获取JSON数据

    Args:
        url (str): URL

    Raises:
        requests.RequestException: Url请求失败
        requests.RequestException: 服务器返回了不正确的答复

    Returns:
        dict: JSON数据
    """
    try:
        resp = requests.get(url, timeout=5)
        resp.raise_for_status()
        resptext = resp.text
    except requests.RequestException as exc:
        raise requests.RequestException("URL请求失败") from exc
    try:
        return json.loads(resptext)
    except json.JSONDecodeError as exc:
        raise requests.RequestException(f"服务器返回了不正确的答复: {resptext}") from exc


class PluginMarket:
    "插件市场类"

    def __init__(self):
        try:
            self.plugins_download_url = Cfg().get_cfg(
                "ToolDelta基本配置.json", {"插件市场源": str})["插件市场源"]
        except Exception:
            self.plugins_download_url = PLUGIN_MARKET_SOURCE_OFFICIAL
        self.plugin_id_name_map: dict = self.get_plugin_id_name_map()

    def enter_plugin_market(self, source_url: str | None = None, in_game=False) -> None:
        """进入插件市场

        Args:
            source_url (str | None, optional): 插件市场源
            in_game (bool, optional): 是否在游戏内
        """
        Print.clean_print("§6正在连接到插件市场..")
        CTXS = 12
        try:
            if isinstance(source_url, str):
                market_datas = self.get_datas_from_market(source_url)
            market_datas = self.get_datas_from_market()
            plugin_ids_map = self.plugin_id_name_map
            plugins_list = list(market_datas["MarketPlugins"].items())
            all_indexes = len(plugins_list)
            now_index = 0
            sum_pages = int((all_indexes - 1) / CTXS) + 1
            now_page = 0
            last_operation = ""
            while True:
                clear_screen()
                Print.print_inf(
                    market_datas["SourceName"] + ": " +
                    market_datas["Greetings"],
                    need_log=False
                )
                now_page = int(now_index / CTXS) + 1
                for i in range(now_index, now_index + CTXS):
                    if i in range(all_indexes):
                        plugin_id = plugins_list[i][0]
                        plugin_name = plugin_ids_map[plugin_id]
                        plugin_basic_datas = plugins_list[i][1]
                        if plugin_basic_datas['plugin-type'] == "classic":
                            plugin_type = "类式"
                        elif plugin_basic_datas['plugin-type'] == "injected":
                            plugin_type = "注入式"
                        else:
                            plugin_type = plugin_basic_datas['plugin-type']
                        Print.print_inf(
                            f" {i + 1}. §e{plugin_name} §av{plugin_basic_datas['version']} §b@{plugin_basic_datas['author']} §d{plugin_type}插件",
                            need_log=False
                        )
                    else:
                        Print.print_inf("")
                Print.print_inf(
                    f"§f第 {now_page} / {sum_pages} 页, 输入 §b+§f/§b- §f翻页", need_log=False)
                Print.print_inf("§f输入插件序号选中插件并查看其下载页", need_log=False)
                last_operation = (
                    (
                        input(Print.fmt_info(
                            "§f回车键继续上次操作, §bq§f退出, 请输入:", "§f 输入 "))
                        or last_operation
                    )
                    .lower()
                    .strip()
                )
                if last_operation == "+":
                    now_index += CTXS
                elif last_operation == "-":
                    now_index -= CTXS
                elif last_operation == "q":
                    break
                else:
                    res = Utils.try_int(last_operation)
                    if res:
                        if res in range(1, all_indexes + 1):
                            plugin_data = self.get_plugin_data_from_market(
                                plugins_list[res - 1][0])
                            ok, pres = self.choice_plugin(plugin_data)
                            if ok:
                                if in_game:
                                    if plugin_data.name not in plugin_group.loaded_plugins_name:
                                        resp = input(
                                            Print.fmt_info(
                                                f"§f可以直接热加载该插件: {plugin_data.name}, 是否加载(§aY§f/§cN§f): ")
                                        ).strip().lower()
                                        if resp == "y":
                                            for i in pres:
                                                if i not in plugin_group.loaded_plugins_name:
                                                    try:
                                                        plugin_group.load_plugin_hot(
                                                            i.name, i.plugin_type)
                                                    except BaseException as err:
                                                        Print.print_err(
                                                            f"插件热加载出现问题: {err}")
                                    else:
                                        Print.print_inf(
                                            "插件已存在, 若要更新版本, 请重启 ToolDelta", need_log=False)
                                        r = input(Print.fmt_info(
                                            "§f输入 §cq §f退出, 其他则返回插件市场"))
                                else:
                                    Print.print_inf(
                                        "下载插件后重启ToolDelta才能生效", need_log=False)
                                r = input(Print.fmt_info(
                                    "§f输入 §cq §f退出, 其他则返回插件市场"))
                                if r.lower() == "q":
                                    break
                            else:
                                Print.print_inf("已取消.", need_log=False)
                                time.sleep(1)
                        else:
                            Print.print_err("超出序号范围")
                if now_index >= all_indexes:
                    now_index = 0
                elif now_index < 0:
                    now_index = max(now_index - CTXS, 0)
        except KeyError as err:
            Print.print_err(f"获取插件市场插件出现问题: 键值对错误: {err}")
            return
        except requests.RequestException as err:
            Print.print_err(f"获取插件市场插件出现问题: {err}")
            return
        except Exception as err:
            Print.print_err("获取插件市场插件出现问题, 报错如下:")
            Print.print_err(traceback.format_exc())
            return
        clear_screen()
        Print.clean_print("§a已从插件市场返回 ToolDelta 控制台.")

    def get_datas_from_market(self, source_url: str | None = None) -> dict:
        """从插件市场获取数据

        Args:
            source_url (str | None, optional): 插件市场源, 默认为None, 有则替换整体插件市场源

        Returns:
            dict: 插件市场数据
        """
        if isinstance(source_url, str):
            self.plugins_download_url = self.plugins_download_url
        market_datas = get_json_from_url(
            url_join(self.plugins_download_url, "market_tree.json")
        )
        return market_datas

    def get_plugin_data_from_market(self, plugin_id: str) -> PluginRegData:
        """从插件市场获取插件数据

        Args:
            plugin_id (str): 插件ID

        Raises:
            KeyError: 无法通过ID查找插件

        Returns:
            PluginRegData: 插件注册数据
        """
        plugin_name = self.plugin_id_name_map.get(plugin_id)
        if plugin_name is None:
            raise KeyError(f"无法通过ID: {plugin_id} 查找插件")
        data_url = self.plugins_download_url + "/" + plugin_name + "/datas.json"
        res = requests.get(data_url, timeout=10)
        res.raise_for_status()
        datas = json.loads(res.text)
        return PluginRegData(plugin_name, datas)

    def choice_plugin(self, plugin_data: PluginRegData) -> tuple[bool, list[PluginRegData]]:
        """选中插件进行介绍与操作

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Returns:
            tuple[bool, list[PluginRegData]]: 是否下载, 下载的插件列表
        """
        pre_plugins_str = (
            ", ".join(
                [f"{k}§7v{v}" for k, v in plugin_data.pre_plugins.items()]) or "无"
        )
        clear_screen()
        Print.print_inf(
            f"{plugin_data.name} v{plugin_data.version_str}", need_log=False
        )
        Print.print_inf(
            f"作者: §f{plugin_data.author}§7, 版本: §f{plugin_data.version_str} §b{plugin_data.plugin_type_str}",
            need_log=False,
        )
        Print.print_inf(f"前置插件: §f{pre_plugins_str}", need_log=False)
        Print.print_inf(f"介绍: {plugin_data.description}", need_log=False)
        Print.print_inf("", need_log=False)
        res = input(Print.fmt_info(
            "§f下载 = §aY§f, 取消 = §cN§f, 请输入:")).lower().strip()
        if res == "y":
            Print.clean_print(f"§6正在下载插件: §f{plugin_data.name}", end="\r")
            pres = self.download_plugin(plugin_data)
            pres.reverse()
            return True, pres
        return False, []

    def get_plugin_id_name_map(self) -> dict:
        """获取插件ID与插件名的映射

        Returns:
            dict: 插件ID与插件名的映射
        """
        res = requests.get(self.plugins_download_url +
                           "/plugin_ids_map.json", timeout=5)
        res.raise_for_status()
        res1: dict = json.loads(res.text)
        self.plugin_id_name_map = res1
        return res1

    def download_plugin(
        self,
        plugin_data: PluginRegData,
        with_pres = True
    ) -> list[PluginRegData]:
        """
        下载插件
        注意: 只能传入由 `get_plugin_data_from_market()` 生成的数据

        Args:
            plugin_data (PluginRegData): 插件注册数据
            with_pres (bool): 是否一同下载前置插件

        Raises:
            ValueError: 未知插件类型

        Returns:
            list[PluginRegData]: 本插件和其前置插件的注册数据列表
        """
        if self.plugin_id_name_map is None:
            self.plugin_id_name_map = self.get_plugin_id_name_map()
        pres = [plugin_data]
        download_paths = self.find_dirs(plugin_data)
        if with_pres:
            for plugin_id in plugin_data.pre_plugins:
                plugin_name = self.plugin_id_name_map[plugin_id]
                Print.clean_print(f"正在下载 {plugin_data.name} 的前置插件 {plugin_name}")
                plugin_datas = self.get_plugin_data_from_market(plugin_id)
                pres += self.download_plugin(plugin_datas)
        cache_dir = tempfile.mkdtemp()
        try:
            for paths in download_paths:
                if not paths.strip():
                    Print.print_war("下载路径为空, 跳过")
                    continue
                url = url_join(self.plugins_download_url, paths)
                # Determine download path based on plugin type
                match plugin_data.plugin_type:
                    case "classic":
                        download_path = os.path.join(
                            "插件文件", TOOLDELTA_CLASSIC_PLUGIN
                        )
                    case "injected":
                        download_path = os.path.join(
                            "插件文件", TOOLDELTA_INJECTED_PLUGIN
                        )
                    case _:
                        raise ValueError(
                            f"未知插件类型: {plugin_data.plugin_type}, 你可能需要通知ToolDelta项目开发组解决"
                        )
                os.makedirs(os.path.join(
                    cache_dir, plugin_data.name), exist_ok=True)
                path_last = path_dir(paths)
                if path_last is not None:
                    # 自动创建文件夹
                    folder_path = os.path.join(cache_dir, path_last)
                    os.makedirs(folder_path, exist_ok=True)
                urlmethod.download_unknown_file(
                    url, os.path.join(cache_dir, paths))
            # Move downloaded files to target download path
            target_path = download_path
            os.makedirs(target_path, exist_ok=True)
            # 制作所需的目录结构
            for root, _, files in os.walk(cache_dir):
                for filename in files:
                    source_file = os.path.join(root, filename)
                    target_file = os.path.join(
                        target_path, os.path.relpath(source_file, cache_dir)
                    )
                    os.makedirs(os.path.dirname(target_file), exist_ok=True)
                    shutil.move(source_file, target_file)
            from .plugin_manager import plugin_manager
            plugin_manager.push_plugin_reg_data(
                self.get_plugin_data_from_market(plugin_data.plugin_id))
            Print.clean_print(f"§a成功下载插件 §f{plugin_data.name}§a 至插件文件夹")
        finally:
            shutil.rmtree(cache_dir)
        return pres

    def find_dirs(self, plugin_data: PluginRegData) -> list[str]:
        """查找插件目录

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Raises:
            KeyError: 目录结构错误
            eExceptionrr: 获取插件市场插件目录结构出现问题

        Returns:
            list[str]: 插件目录列表
        """
        try:
            data = get_json_from_url(
                url_join(self.plugins_download_url, "directory.json")
            )
            data_list = []
            for folder, files in data.items():
                if plugin_data.name == folder.split("/")[0]:
                    # 展开
                    for file in files:
                        data_list.append(folder + r"/" + file)
            return data_list
        except KeyError as err:
            Print.print_err(f"获取插件市场插件目录结构出现问题: 无法找到 {err}, 有可能是未来得及更新目录")
            raise KeyError(f"无法找到 {err}, 有可能是未来得及更新目录") from err
        except Exception as err:
            Print.print_err(f"获取插件市场插件目录结构出现问题: {err}")
            raise KeyError(f"获取插件市场插件目录结构出现问题: {err}")from err

    def get_latest_plugin_version(self, plugin_id: str) -> str:
        """获取最新插件版本

        Args:
            plugin_id (str): 插件ID

        Raises:
            KeyError: 无法通过ID获取最新插件版本

        Returns:
            str: 最新插件版本
        """
        result = get_json_from_url(
            url_join(self.plugins_download_url, "latest_versions.json")
        ).get(plugin_id)
        if isinstance(result, str):
            return result
        raise KeyError(f"无法通过ID: {plugin_id} 获取最新插件版本")


market = PluginMarket()



"""
## ToolDelta: NEMC Rental Server R-Plugin Loader ##
ToolDelta: 网易我的世界手机版 租赁服机器人式插件加载器
 - 进入ToolDelta界面:\n
    from tooldelta import client_title\n
    client_title()
 - 快速启动ToolDelta:\n
    from tooldelta import start_tool_delta\n
    start_tool_delta()
"""
from .color_print import Print
from .plugin_load.PluginGroup import Plugin
from .starter import plugin_group as plugins
from .starter import start_tool_delta, safe_jump
from .starter import frame as tooldelta
from .frame import Config, Utils, Frame, GameCtrl
from .launch_options import client_title

# 重定向 Builtins
Builtins = Utils


"ToolDelta类式插件"
import importlib
import os
import sys
import traceback
import zipfile
from typing import TYPE_CHECKING, Union, TypeVar
from ...color_print import Print
from ...utils import Utils
from ...cfg import Cfg
from ...plugin_load import plugin_is_enabled, NotValidPluginError
from ...constants import TOOLDELTA_CLASSIC_PLUGIN

if TYPE_CHECKING:
    # 类型注释
    from ...frame import Frame
    from ...plugin_load.PluginGroup import PluginGroup

__caches__ = {
    "plugin": None,
    "api_name": "",
    "frame": None
}

class Plugin:
    "插件信息主类"
    name: str = ""
    version = (0, 0, 1)
    author = "?"
    description = "..."

    def __init__(self, frame: "Frame"):
        self.frame = frame
        self.game_ctrl = frame.get_game_control()

    @property
    def data_path(self) -> str:
        "该插件的数据文件夹路径 (调用时直接创建数据文件夹)"
        path = os.path.join("插件数据文件", self.name)
        os.makedirs(path, exist_ok=True)
        return path

_PLUGIN_CLS_TYPE = TypeVar("_PLUGIN_CLS_TYPE")

def add_plugin(plugin: type[_PLUGIN_CLS_TYPE]) -> type[_PLUGIN_CLS_TYPE]:
    """添加ToolDelta类式插件的插件主类

    Args:
        plugin (type[Plugin]): 插件主类

    Raises:
        NotValidPluginError: 插件主类必须继承Plugin类

    Returns:
        type[Plugin]: 插件主类
    """
    try:
        if not Plugin.__subclasscheck__(plugin):
            raise NotValidPluginError(f"插件主类必须继承Plugin类 而不是 {plugin}")
    except TypeError as exc:
        raise NotValidPluginError(
            f"插件主类必须继承Plugin类 而不是 {plugin.__class__}") from exc
    if __caches__["plugin"] is not None:
        raise NotValidPluginError(
            "调用了多次 @add_plugin"
        )
    plugin_ins = plugin(__caches__["frame"]) # type: ignore
    __caches__["plugin"] = plugin_ins
    return plugin

def add_plugin_as_api(apiName: str):
    """添加ToolDelta类式插件主类, 同时作为API插件提供接口供其他插件进行使用

    Args:
        apiName (str): API名
    """
    def _add_plugin_2_api(api_plugin: type[_PLUGIN_CLS_TYPE]) -> type[_PLUGIN_CLS_TYPE]:
        if not Plugin.__subclasscheck__(api_plugin):
            raise NotValidPluginError("API插件主类必须继承Plugin类")
        if __caches__["plugin"] is not None:
            raise NotValidPluginError(
                "调用了多次 @add_plugin"
            )
        plugin_ins = api_plugin(__caches__["frame"]) # type: ignore
        __caches__["plugin"] = plugin_ins
        __caches__["api_name"] = apiName
        return api_plugin

    return _add_plugin_2_api


# Plugin get and execute

def read_plugins(plugin_grp: "PluginGroup") -> None:
    """读取插件

    Args:
        plugin_grp (PluginGroup): 插件组
    """
    PLUGIN_PATH = os.path.join("插件文件", TOOLDELTA_CLASSIC_PLUGIN)
    sys.path.append(os.path.join("插件文件", TOOLDELTA_CLASSIC_PLUGIN))
    for plugin_dir in os.listdir(PLUGIN_PATH):
        if not plugin_is_enabled(plugin_dir):
            continue
        if (
            not os.path.isdir(os.path.join(
                PLUGIN_PATH, plugin_dir.strip(".zip")))
            and os.path.isfile(os.path.join(PLUGIN_PATH, plugin_dir))
            and plugin_dir.endswith(".zip")
        ):
            Print.print_with_info(f"§6正在解压插件{plugin_dir}, 请稍后", "§6 解压 ")
            _unzip_plugin(
                os.path.join(PLUGIN_PATH, plugin_dir),
                os.path.join(PLUGIN_PATH, plugin_dir.strip(".zip")),
            )
            Print.print_suc(f"§a成功解压插件{plugin_dir} -> 插件目录")
            plugin_dir = plugin_dir.strip(".zip")
        if os.path.isdir(os.path.join(PLUGIN_PATH, plugin_dir)):
            sys.path.append(os.path.join(PLUGIN_PATH, plugin_dir))
            load_plugin(plugin_grp, plugin_dir)
            plugin_grp.loaded_plugins_name.append(plugin_dir)


def load_plugin(plugin_group: "PluginGroup", plugin_dirname: str) -> Union[None, Plugin]:
    """加载插件

    Args:
        plugin_group (PluginGroup): 插件组类
        plugin_dirname (str): 插件目录名
        hot_load (bool, optional): 是否热加载

    Raises:
        ValueError: 插件组未初始化读取
        ValueError: 插件组未绑定框架
        ValueError: 插件主类需要作者名
        ValueError: 插件主类需要作者名
        NotValidPluginError: 插件 不合法
        SystemExit: 插件名字不合法
        SystemExit: 插件配置文件报错
        SystemExit: 插件读取数据失败

    Returns:
        Union[None, Plugin]: 插件实例
    """
    if isinstance(plugin_group, type(None)):
        raise ValueError("插件组未初始化读取")
    if isinstance(plugin_group.linked_frame, type(None)):
        raise ValueError("插件组未绑定框架")
    plugin_group.plugin_added_cache["packets"].clear()
    plugin_group.broadcast_evts_cache.clear()
    __caches__["plugin"] = None
    __caches__["api_name"] = ""
    try:
        if os.path.isfile(
            os.path.join(
                "插件文件", TOOLDELTA_CLASSIC_PLUGIN, plugin_dirname, "__init__.py"
            )
        ):
            importlib.import_module(plugin_dirname)
        else:
            Print.print_war(f"{plugin_dirname} 文件夹 未发现插件文件, 跳过加载")
            return
        plugin_or_none : Plugin | None = __caches__.get("plugin")
        if plugin_or_none is None:
            raise NotValidPluginError(
                "需要调用1次 @plugins.add_plugin 以注册插件主类, 然而没有调用"
            )
        plugin: Plugin = plugin_or_none
        if plugin.name is None or plugin.name == "":
            raise ValueError(f"插件主类 {plugin.__class__.__name__} 需要作者名")
        plugin_group.plugins.append(plugin)
        _v0, _v1, _v2 = plugin.version
        for evt_name in (
            "on_def",
            "on_inject",
            "on_player_prejoin",
            "on_player_join",
            "on_player_message",
            "on_player_death",
            "on_player_leave",
            "on_frame_exit"
        ):
            if hasattr(plugin, evt_name):
                plugin_group.plugins_funcs[evt_name].append(
                    [plugin.name, getattr(plugin, evt_name)]
                )
        Print.print_suc(
            f"成功载入插件 {plugin.name} 版本: {_v0}.{_v1}.{_v2} 作者：{plugin.author}"
        )
        plugin_group.normal_plugin_loaded_num += 1
        if plugin_group.plugin_added_cache["packets"] != []:
            for pktType, func in plugin_group.plugin_added_cache["packets"]:
                ins_func = getattr(plugin, func.__name__)
                if ins_func is None:
                    raise NotValidPluginError("数据包监听不能在主插件类以外定义")
                plugin_group._add_listen_packet_id(pktType)
                plugin_group._add_listen_packet_func(
                    pktType, ins_func
                )
        if __caches__["api_name"] != "":
            plugin_group.plugins_api[__caches__["api_name"]] = plugin
        if plugin_group.broadcast_evts_cache != {}:
            for evt, funcs in plugin_group.broadcast_evts_cache.items():
                for func in funcs:
                    ins_func = getattr(plugin, func.__name__)
                    if ins_func is None:
                        raise NotValidPluginError("广播事件监听不能在主插件类以外定义")
                    plugin_group._add_broadcast_evt(evt, ins_func)
        return plugin
    except NotValidPluginError as err:
        Print.print_err(f"插件 {plugin_dirname} 不合法: {err.args[0]}")
        raise SystemExit from err
    except Cfg.ConfigError as err:
        Print.print_err(f"插件 {plugin_dirname} 配置文件报错：{err}")
        Print.print_err("你也可以直接删除配置文件, 重新启动ToolDelta以自动生成配置文件")
        raise SystemExit from err
    except Utils.SimpleJsonDataReader.DataReadError as err:
        Print.print_err(f"插件 {plugin_dirname} 读取数据失败: {err}")
    except plugin_group.linked_frame.SystemVersionException as err:
        Print.print_err(f"插件 {plugin_dirname} 需要更高版本的ToolDelta加载: {err}")
    except Exception as err:
        Print.print_err(f"加载插件 {plugin_dirname} 出现问题, 报错如下: ")
        Print.print_err("§c" + traceback.format_exc())
        raise SystemExit from err
    return None

def _unzip_plugin(zip_dir: str, exp_dir: str) -> None:
    """解压插件ZIP包

    Args:
        zip_dir (str): 压缩文件路径
        exp_dir (str): 解压目录
    """
    try:
        f = zipfile.ZipFile(zip_dir, "r")
        f.extractall(exp_dir)
    except Exception as err:
        Print.print_err(f"zipfile: 解压失败: {err}")
        raise EOFError("解压失败") from err

def _init_frame(frame: "Frame"):
    __caches__["frame"] = frame


"实用函数"
import zipfile
from typing import TYPE_CHECKING
from ..color_print import Print
from ..packets import Packet_CommandOutput

if TYPE_CHECKING:
    from tooldelta import Frame, GameCtrl

game_ctrl: "GameCtrl" = None # type: ignore

__all__ = [
    "getTarget",
    "getPos",
    "getItem",
    "getScore",
    "isCmdSuccess"
]

# set_frame

def check_gamectrl_avali():
    "检查GameCtrl是否可用"
    if game_ctrl is None:
        raise ValueError("GameControl 不可用")

def set_frame(frame: "Frame"):
    "载入系统框架"
    global game_ctrl
    game_ctrl = frame.get_game_control()

# utils

def getTarget(sth: str, timeout: bool | int = 5) -> list:
    """
    获取符合目标选择器实体的列表

    参数:
        sth: 目标选择器
        timeout: 超时时间，默认为5秒
    异常:
        ValueError: 指令返回超时, 或者无法获取目标
    """
    check_gamectrl_avali()
    if not sth.startswith("@"):
        raise ValueError("我的世界目标选择器格式错误(getTarget必须使用目标选择器)")
    result = game_ctrl.sendwscmd(f"/testfor {sth}", True, timeout)
    if result is None:
        raise ValueError("获取目标失败")
    result = (
        result
        .OutputMessages[0]
        .Parameters
    )
    if result:
        result = result[0]
        return result.split(", ")
    raise ValueError("获取目标失败")

def getPos(targetNameToGet: str, timeout: float | int = 5) -> dict:
    """获取目标玩家的位置信息

    参数:
        targetNameToGet: 目标玩家的名称
        timeout: 超时时间（秒）。默认为5秒

    异常:
        ValueError: 当目标玩家不存在时抛出该异常
        ValueError: 当获取位置信息失败时抛出该异常
        AttributeError: 当获取玩家UUID失败时抛出该异常
    """
    check_gamectrl_avali()
    if targetNameToGet not in game_ctrl.allplayers or targetNameToGet.startswith("@"):
        raise ValueError(f"Player {targetNameToGet} does not exist.")
    result = game_ctrl.sendwscmd(
        f'/querytarget @a[name="{targetNameToGet}"]', True, timeout)
    if result is None:
        raise ValueError("Failed to get the position.")
    if not result.OutputMessages[0].Success:
        raise ValueError(
            f"Failed to get the position: {result.OutputMessages[0]}")
    parameter = result.OutputMessages[0].Parameters[0]
    if isinstance(parameter, str):
        raise ValueError("Failed to get the position.")
    result = {}

    if game_ctrl.players_uuid is None:
        raise AttributeError("Failed to get the players_uuid.")
    targetName = targetNameToGet
    x = (
        parameter[0]["position"]["x"]
        if parameter[0]["position"]["x"] >= 0
        else parameter[0]["position"]["x"] - 1
    )
    y = parameter[0]["position"]["y"] - 1.6200103759765
    z = (
        parameter[0]["position"]["z"]
        if parameter[0]["position"]["z"] >= 0
        else parameter[0]["position"]["z"] - 1
    )
    position = {
        "x": float(f"{x:.2f}"),
        "y": float(f"{y:.2f}"),
        "z": float(f"{z:.2f}"),
    }
    dimension = parameter[0]["dimension"]
    yRot = parameter[0]["yRot"]
    result[targetName] = {
        "dimension": dimension,
        "position": position,
        "yRot": yRot,
    }
    if targetNameToGet == "@a":
        return result
    if len(result) != 1:
        raise ValueError("Failed to get the position.")
    if targetNameToGet.startswith("@a"):
        return list(result.values())[0]
    return result[targetNameToGet]

def getItem(targetName: str, itemName: str, itemSpecialID: int = -1) -> int:
    """
    获取玩家背包内指定的物品的数量
    参数:
        targetName (str): 玩家选择器 / 玩家名
        itemName (str): 物品ID
        itemSpecialID (int) = -1: 物品特殊值
    """
    if (
        (targetName not in game_ctrl.allplayers)
        and (targetName != game_ctrl.bot_name)
        and (not targetName.startswith("@a"))
    ):
        raise Exception("未找到目标玩家")
    result: Packet_CommandOutput = game_ctrl.sendwscmd(
        f"/clear {targetName} {itemName} {itemSpecialID} 0", True
    ) # type: ignore
    if result.OutputMessages[0].Message == "commands.generic.syntax":
        raise Exception("物品ID错误")
    if result.OutputMessages[0].Message == "commands.clear.failure.no.items":
        return 0
    return int(result.OutputMessages[0].Parameters[1])

def getPosXYZ(player, timeout=30) -> tuple[float, float, float]:
    """
    获取玩家的简略坐标值, 并以坐标三元元组返回
    参数:
        player (str): 玩家名
        timeout (int): 最长超时时间
    返回:
        tuple[float, float, float]
    """
    res = getPos(player, timeout=timeout)["position"]
    return res["x"], res["y"], res["z"]

def getScore(scoreboardNameToGet: str, targetNameToGet: str) -> int | dict:
    """
    获取计分板分数
    参数:
        scoreboardNameToGet: 计分板名
        targetNameToGet: 获取分数的对象/目标选择器
    返回:
        分数: int
    异常:
        ValueError: 无法获取分数
    """
    if scoreboardNameToGet == "*":
        raise ValueError("暂不能使用 * 作为计分板参数")
    resultList = game_ctrl.sendwscmd(
        f"/scoreboard players list {targetNameToGet}", True
    ).OutputMessages # type: ignore
    result = {}
    result2 = {}
    for i in resultList:
        Message = i.Message
        if Message == r"commands.scoreboard.players.list.player.empty":
            continue
        if Message == r"§a%commands.scoreboard.players.list.player.count":
            targetName = i.Parameters[1][1:]
        elif Message == "commands.scoreboard.players.list.player.entry":
            if targetName == "commands.scoreboard.players.offlinePlayerName":
                continue
            scoreboardName = i.Parameters[2]
            targetScore = int(i.Parameters[0])
            if targetName not in result:
                result[targetName] = {}
            result[targetName][scoreboardName] = targetScore
            if scoreboardName not in result2:
                result2[scoreboardName] = {}
            result2[scoreboardName][targetName] = targetScore
    if not (result or result2):
        raise Exception("获取计分板分数失败")
    try:
        if targetNameToGet == "*" or targetNameToGet.startswith("@"):
            return result2[scoreboardNameToGet]
        if scoreboardNameToGet == "*":
            return result[targetNameToGet]
        return result[targetNameToGet][scoreboardNameToGet]
    except KeyError as err:
        raise Exception(f"获取计分板分数失败: {err}")

def isCmdSuccess(cmd: str, timeout=30):
    """
    获取命令执行成功与否的状态
    参数:
        cmd: MC指令
        timeout: 超时时间
    返回:
        命令执行是否成功: bool
    """
    res = game_ctrl.sendwscmd(cmd, True, timeout).SuccessCount # type: ignore
    return bool(res)